import tempfile
import unittest
from glob import glob
from pathlib import Path

from tirex_tracker import ExportFormat, clear_file_register, register_file, tracking


class TestRegisteredMetadata(unittest.TestCase):
    def test_without_registered_value(self) -> None:
        expected = ["metadata.yml"]
        clear_file_register()

        with tempfile.TemporaryDirectory() as tmp:
            with tracking(export_file_path=Path(tmp) / "metadata.yml", export_format=ExportFormat.IR_METADATA):
                print("hello world")
            actual = [i.split("/")[-1] for i in glob(f"{tmp}/*")]

        self.assertEqual(expected, actual)

    def test_with_registered_value(self) -> None:
        expected = set(["metadata.yml", "test_integration.py"])
        clear_file_register()
        register_file(Path(__file__).parent, Path("test_integration.py"))

        with tempfile.TemporaryDirectory() as tmp:
            with tracking(export_file_path=Path(tmp) / "metadata.yml", export_format=ExportFormat.IR_METADATA):
                print("hello world")
            actual = set([i.split("/")[-1] for i in glob(f"{tmp}/*")])

        self.assertEqual(expected, actual)

    def test_with_registered_value_to_subdir(self) -> None:
        expected = set(["test_integration.py"])
        clear_file_register()
        register_file(Path(__file__).parent, Path("test_integration.py"), "subdir")

        with tempfile.TemporaryDirectory() as tmp:
            with tracking(export_file_path=Path(tmp) / "metadata.yml", export_format=ExportFormat.IR_METADATA):
                print("hello world")
            actual = set([i.split("/")[-1] for i in glob(f"{tmp}/subdir/*")])

        self.assertEqual(expected, actual)



import tempfile
import unittest
from glob import glob
from pathlib import Path

from tirex_tracker import ExportFormat, clear_file_register, tracking


class TestRedundantTracking(unittest.TestCase):
    def test_with_different_metadata_files(self) -> None:
        expected = ["metadata1.yml", "metadata2.yml"]
        clear_file_register()

        with tempfile.TemporaryDirectory() as tmp:
            with tracking(export_file_path=Path(tmp) / "metadata1.yml", export_format=ExportFormat.IR_METADATA):
                print("hello world")

            with tracking(export_file_path=Path(tmp) / "metadata2.yml", export_format=ExportFormat.IR_METADATA):
                print("hello world")

            actual = sorted([i.split("/")[-1] for i in glob(f"{tmp}/*")])

        self.assertEqual(expected, actual)

    def test_with_identical_metadata_files(self) -> None:
        expected = ["metadata.yml"]
        clear_file_register()

        with tempfile.TemporaryDirectory() as tmp:
            with tracking(export_file_path=Path(tmp) / "metadata.yml", export_format=ExportFormat.IR_METADATA):
                print("hello world")

            with self.assertRaises(ValueError): # noqa: SIM117
                with tracking(export_file_path=Path(tmp) / "metadata.yml", export_format=ExportFormat.IR_METADATA):
                    print("hello world")

            actual = sorted([i.split("/")[-1] for i in glob(f"{tmp}/*")])

        self.assertEqual(expected, actual)



"""
A short note on how we perform error handling.
What we want to achieve: The following code should abort execution of do_something() if an error occurs during the
tracking:

```py
with tracking():
    # do something
```

This is difficult since all tracking happens in a separate thread. This means that we can not simply throw an exception
as this exception would only be handled in the local thread. The solution is that the error callback from TIREx (which
is run within the tracking thread when an error occurs) signals the monitored thread. This interrupts do_something()
asynchronously. We must have added a signal handler before do_something() runs which in turn throws an exception. That
is, for error handling, the following steps are necessary:
1. Instate signal handler (__instate_abort_signal_handler)
2. Set the abort callback to signal the tracked thread (__instate_abort_callback)
3. Start tracking
4. Perform the work...
5. Deinitialize the signal handler and callback (deinit_error_handling)

Step 1 & 2 are combined in init_error_handling.
"""

import os
import signal
from ctypes import CFUNCTYPE, c_char_p
from types import FrameType
from typing import Any, List, Optional, Tuple, Union

from .constants import ENCODING
from .library import LIBRARY

__SIGNALS_TO_TRY = [signal.Signals.SIGUSR1, signal.Signals.SIGUSR2]
ABORT_HANDLE = Tuple[signal.Signals, Any]


def __instate_abort_signal_handler(handler: "signal._HANDLER") -> Optional[signal.Signals]:
    for sig in __SIGNALS_TO_TRY:
        if signal.getsignal(sig) not in (signal.SIG_IGN, signal.SIG_DFL, None):
            continue  # Signal already in use by someone else
        signal.signal(sig, handler)
        return sig
    return None


def __instate_abort_callback(sig: signal.Signals, strref: List[str]) -> Any:
    pid = os.getpid()

    @CFUNCTYPE(None, c_char_p)
    def __abort_callback(message: bytes) -> None:
        strref[0] = message.decode(ENCODING)
        # os.kill does seem weird here but it actually only sends a signal instead of killing the thread:
        # https://docs.python.org/3/library/os.html#os.kill: "Send signal sig to process pid".
        os.kill(pid, sig)

    LIBRARY.tirexSetAbortCallback(__abort_callback)
    # We need to return __abort_callback here since we need to keep it referenced as long as it is set as the abort
    # callback.
    return __abort_callback


def init_error_handling() -> ABORT_HANDLE:
    strref = [""]

    def sig_handler(signum: int, frame: Union[FrameType, None]) -> None:
        raise RuntimeError(f"TIREx Tracker encountered a problem: {strref[0]}")

    sig = __instate_abort_signal_handler(sig_handler)
    if sig is None:
        raise RuntimeError(
            f"Failed to instantiate a signal handler, all tested signals ({__SIGNALS_TO_TRY}) are already in use"
        )
    return sig, __instate_abort_callback(sig, strref)


def deinit_error_handling(handle: ABORT_HANDLE):
    sig, _ = handle
    signal.signal(sig, signal.SIG_DFL)
    LIBRARY.tirexSetAbortCallback(None)



import tempfile
import unittest
from pathlib import Path

import yaml

from tirex_tracker import ExportFormat, clear_metadata_register, register_metadata, tracking


class TestRegisteredMetadata(unittest.TestCase):
    def test_without_registered_value(self) -> None:
        clear_metadata_register()
        with tempfile.TemporaryDirectory() as tmp:
            result_path = Path(tmp) / "example-metadata.yml"
            with tracking(export_file_path=result_path, export_format=ExportFormat.IR_METADATA):
                print("hello world")

            actual = yaml.safe_load(result_path.read_text())
            self.assertIn("implementation", actual)
            self.assertNotIn("data", actual)

    def test_with_registered_value(self) -> None:
        clear_metadata_register()
        register_metadata({"data": {"test collection": {"name": "some-dataset"}}})
        with tempfile.TemporaryDirectory() as tmp:
            result_path = Path(tmp) / "example-metadata.yml"
            with tracking(export_file_path=result_path, export_format=ExportFormat.IR_METADATA):
                print("hello world")

            actual = yaml.safe_load(result_path.read_text())
            self.assertIn("implementation", actual)
            self.assertIn("data", actual)



from __future__ import annotations

from ctypes import c_char_p, c_wchar_p, cast
from typing import TYPE_CHECKING

from tirex_tracker import Measure, ResultEntry, ResultType
from tirex_tracker import _ResultEntry as NativeResultEntry

if TYPE_CHECKING:
    from typing import Any, Callable, Dict


_native_result_entry_parsers: Dict[ResultType, Callable[[int], Any]] = {
    ResultType.STRING: lambda ptr: cast(ptr, c_char_p).value,
    ResultType.INTEGER: lambda ptr: NotImplemented,  # TODO: implement
    ResultType.FLOATING: lambda ptr: NotImplemented,  # TODO: implement
    ResultType.BOOLEAN: lambda ptr: NotImplemented,  # TODO: implement
    ResultType.WSTRING: lambda ptr: cast(ptr, c_wchar_p).value,
    ResultType.STRING_LIST: lambda ptr: NotImplemented,  # TODO: implement
    ResultType.INTEGER_LIST: lambda ptr: NotImplemented,  # TODO: implement
    ResultType.FLOATING_LIST: lambda ptr: NotImplemented,  # TODO: implement
    ResultType.BOOLEAN_LIST: lambda ptr: NotImplemented,  # TODO: implement
    ResultType.WSTRING_LIST: lambda ptr: NotImplemented,  # TODO: implement
}


def parse_native_result_entry(entry: NativeResultEntry) -> ResultEntry:
    rtype = ResultType(entry.type)
    value = _native_result_entry_parsers[rtype](entry.value)
    return ResultEntry(source=Measure(entry.source), value=value, type=rtype)



from __future__ import annotations

from collections import defaultdict
from ctypes import CFUNCTYPE, c_char_p, c_int, c_size_t, pointer
from dataclasses import dataclass
from enum import Enum
from functools import wraps
from gzip import open as gzip_open
from io import BytesIO
from json import dumps, loads
from pathlib import Path
from shutil import copy, rmtree
from sys import argv, executable, version_info
from sys import modules as sys_modules
from traceback import extract_stack
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    ContextManager,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
    Union,
    ValuesView,
    cast,
    overload,
)

from importlib_metadata import distributions
from IPython.core.getipython import get_ipython
from typing_extensions import ParamSpec, Self, TypeAlias  # type: ignore
from yaml import safe_dump as yaml_safe_dump
from yaml import safe_load as yaml_safe_load

from ._utils.constants import ALL_AGGREGATIONS, ALL_MEASURES, Aggregation, Error, LogLevel, Measure, ResultType
from ._utils.constants import ENCODING as _ENCODING
from ._utils.errorhandling import ABORT_HANDLE as _ABORT_HANDLE
from ._utils.errorhandling import deinit_error_handling as _deinit_error_handling
from ._utils.errorhandling import init_error_handling as _init_error_handling
from ._utils.library import (
    _PYTHON_MEASURES,
    MeasureInfo,
    ProviderInfo,
    ResultEntry,
    _MeasureConfiguration,
    _MeasureInfo,
    _ProviderInfo,
    _Result,
    _ResultEntry,
    _TrackingHandle,
    provider_version,
)
from ._utils.library import LIBRARY as _LIBRARY
from ._utils.library import (
    NULL_MEASURE_CONFIGURATION as _NULL_MEASURE_CONFIGURATION,
)
from .archive_utils import create_code_archive, git_repo_or_none

if TYPE_CHECKING:
    from ctypes import Array
    from ctypes import _CFunctionType as CFunctionType  # type: ignore
    from ctypes import _Pointer as Pointer  # type: ignore

PathLike: TypeAlias = Optional[Union[str, Path]]

P = ParamSpec("P")
T = TypeVar("T")

_REGISTERED_METADATA: MutableMapping[str, Any] = {}
_REGISTERED_FILES: List[Tuple[Path, Path, str]] = []


class ResultsAccessor(Protocol):
    results: Mapping[Measure, ResultEntry]


class LogCallback(Protocol):
    def __call__(self, level: LogLevel, component: str, message: str) -> None:
        pass


_PYTHON_PROVIDER = ProviderInfo(
    name="Python",
    description="Python-specific measures.",
    version=provider_version,
)


def _python_result_entry(measure: Measure, value: Any) -> ResultEntry:
    return ResultEntry(
        source=measure,
        value=dumps(value),
        type=_PYTHON_MEASURES[measure].data_type,
    )


def _add_python_result_entry(
    results: MutableMapping[Measure, ResultEntry],
    measure: Measure,
    measures: Iterable[Measure],
    value: Any,
) -> None:
    if measure not in measures:
        return
    results[measure] = _python_result_entry(measure, value)


def _get_python_info(
    measures: Iterable[Measure], export_file_path: Optional[PathLike] = None
) -> Tuple[Mapping[Measure, ResultEntry], Iterable[Measure]]:
    results: MutableMapping[Measure, ResultEntry] = {}

    _add_python_result_entry(
        results=results,
        measure=Measure.PYTHON_VERSION,
        measures=measures,
        value=f"{version_info.major}.{version_info.minor}.{version_info.micro}",
    )
    _add_python_result_entry(
        results=results,
        measure=Measure.PYTHON_EXECUTABLE,
        measures=measures,
        value=executable,
    )
    _add_python_result_entry(
        results=results,
        measure=Measure.PYTHON_ARGUMENTS,
        measures=measures,
        value=argv,
    )
    modules = sorted({module.split(".")[0] for module in sys_modules if not module.startswith("_")})
    _add_python_result_entry(
        results=results,
        measure=Measure.PYTHON_MODULES,
        measures=measures,
        value=modules,
    )
    installed_packages = sorted({f"{distribution.name}=={distribution.version}" for distribution in distributions()})
    _add_python_result_entry(
        results=results,
        measure=Measure.PYTHON_INSTALLED_PACKAGES,
        measures=measures,
        value=installed_packages,
    )

    ipython = get_ipython()
    is_interactive = ipython is not None
    _add_python_result_entry(
        results=results,
        measure=Measure.PYTHON_IS_INTERACTIVE,
        measures=measures,
        value=is_interactive,
    )

    if ipython is None:
        script_file_path = Path(extract_stack()[0].filename).resolve()
        repo = git_repo_or_none(script_file_path)
        if repo is not None and repo.working_tree_dir is not None:
            script_file_path = script_file_path.relative_to(repo.working_tree_dir)
        _add_python_result_entry(
            results=results,
            measure=Measure.PYTHON_SCRIPT_FILE_PATH,
            measures=measures,
            value=str(script_file_path),
        )

    if export_file_path is not None:
        # Create a utility directory as a sibling to the export file, containing the code archive.
        metadata_directory_path = Path(export_file_path).parent / ".tirex-tracker"

        # Clear and create the directory.
        if metadata_directory_path.exists():
            rmtree(metadata_directory_path)
        metadata_directory_path.mkdir(parents=True)

        archive_paths = create_code_archive(metadata_directory_path)

        _add_python_result_entry(
            results=results,
            measure=Measure.PYTHON_CODE_ARCHIVE_PATH,
            measures=measures,
            value=str(archive_paths.zip_file_path),
        )
        _add_python_result_entry(
            results=results,
            measure=Measure.PYTHON_SCRIPT_FILE_PATH_IN_CODE_ARCHIVE,
            measures=measures,
            value=str(archive_paths.script_file_path_in_zip),
        )
        if archive_paths.notebook_file_path_in_zip is not None:
            _add_python_result_entry(
                results=results,
                measure=Measure.PYTHON_NOTEBOOK_FILE_PATH_IN_CODE_ARCHIVE,
                measures=measures,
                value=str(archive_paths.notebook_file_path_in_zip),
            )

    measures = {measure for measure in measures if measure not in _PYTHON_MEASURES}

    return results, measures


def _deep_merge(dict1: dict, dict2: dict) -> dict:
    for key in dict2:
        if key in dict1 and isinstance(dict1[key], dict) and isinstance(dict2[key], dict):
            _deep_merge(dict1[key], dict2[key])
        else:
            dict1[key] = dict2[key]
    return dict1


def _recursive_defaultdict() -> dict:
    return defaultdict(_recursive_defaultdict)


def _recursive_undefaultdict(dict: dict) -> dict:
    return {
        key: (_recursive_undefaultdict(value) if isinstance(value, defaultdict) else value)
        for key, value in dict.items()
    }


class ExportFormat(Enum):
    IR_METADATA = "ir_metadata"


def _handle_error(error_int: int) -> None:
    error = Error(error_int)
    if error == Error.SUCCESS:
        return
    elif error == Error.INVALID_ARGUMENT:
        raise ValueError("Invalid argument in native call.")
    raise RuntimeError(f"An error occured: {error_int}")


__callback: "Optional[CFunctionType]"


def set_log_callback(log_callback: Optional[LogCallback] = None) -> None:
    def _to_native_log_callback(log_callback: LogCallback):
        @CFUNCTYPE(None, c_int, c_char_p, c_char_p)
        def _log_callback(level: int, component: bytes, message: bytes) -> None:
            log_callback(LogLevel(level), component.decode(_ENCODING), message.decode(_ENCODING))

        return _log_callback

    global __callback  # We need to store a reference to the callback to avoid deletion through GC
    __callback = _to_native_log_callback(log_callback) if log_callback is not None else None

    _LIBRARY.tirexSetLogCallback(__callback)


def provider_infos() -> Collection[ProviderInfo]:
    num_providers = _LIBRARY.tirexDataProviderGetAll((_ProviderInfo * 0)(), 0)
    if num_providers == 0:
        return []
    providers: "Array[_ProviderInfo]" = (_ProviderInfo * num_providers)()
    _LIBRARY.tirexDataProviderGetAll(providers, num_providers)
    providers_iterable: Iterable[_ProviderInfo] = cast(Iterable[_ProviderInfo], providers)
    return [provider.to_provider() for provider in providers_iterable] + [_PYTHON_PROVIDER]


def measure_infos() -> Mapping[Measure, MeasureInfo]:
    measure_infos: MutableMapping[Measure, MeasureInfo] = {}
    for measure in ALL_MEASURES:
        if measure in _PYTHON_MEASURES:
            measure_infos[measure] = _PYTHON_MEASURES[measure]
            continue
        measure_info_pointer = pointer(pointer(_MeasureInfo()))
        error_int = _LIBRARY.tirexMeasureInfoGet(measure.value, measure_info_pointer)
        _handle_error(error_int)
        measure_info = measure_info_pointer.contents.contents
        measure_infos[measure] = measure_info.to_measure_info()
    return measure_infos


def _parse_results(result: "Pointer[_Result]") -> Mapping[Measure, ResultEntry]:
    from tirex_tracker._utils.results import parse_native_result_entry  # avoid circular imports

    num_entries_pointer = pointer(c_size_t())
    error_int = _LIBRARY.tirexResultEntryNum(result, num_entries_pointer)
    _handle_error(error_int)
    num_entries = num_entries_pointer.contents.value
    entries: List[ResultEntry] = []
    for index in range(num_entries):
        entry_pointer = pointer(_ResultEntry())
        error_int = _LIBRARY.tirexResultEntryGetByIndex(result, c_size_t(index), entry_pointer)
        _handle_error(error_int)
        entries.append(parse_native_result_entry(entry_pointer.contents))
    _LIBRARY.tirexResultFree(result)
    results = {entry.source: entry for entry in entries}
    return results


def _prepare_measure_configurations(
    measures: Iterable[Measure],
) -> "Array[_MeasureConfiguration]":
    configs = [_MeasureConfiguration(measure.value, Aggregation.NO.value) for measure in measures] + [
        _NULL_MEASURE_CONFIGURATION
    ]
    configs_array = (_MeasureConfiguration * (len(configs)))(*configs)
    return configs_array


# TODO: Add aggregation(s) (mapping) parameter.
def fetch_info(
    measures: Iterable[Measure] = ALL_MEASURES,
) -> Mapping[Measure, ResultEntry]:
    # Get Python info first, and then strip Python measures from the list.
    python_info, remaining_measures = _get_python_info(measures)

    # Prepare the measure configurations.
    configs_array = _prepare_measure_configurations(remaining_measures)

    result_pointer = pointer(pointer(_Result()))
    error_int = _LIBRARY.tirexFetchInfo(configs_array, result_pointer)
    _handle_error(error_int)

    return {
        **_parse_results(result_pointer.contents),
        **python_info,
    }


@dataclass(frozen=True)
class TrackingHandle(ContextManager["TrackingHandle"], Mapping[Measure, ResultEntry]):
    _fetch_info_result: "Pointer[_Result]"
    _error_handle: _ABORT_HANDLE
    _tracking_handle: "Pointer[_TrackingHandle]"
    _python_info: Mapping[Measure, ResultEntry]
    _system_name: Optional[str]
    _system_description: Optional[str]
    _export_file_path: Optional[PathLike]
    _export_format: Optional[ExportFormat]
    results: MutableMapping[Measure, ResultEntry]

    # TODO: Add aggregation(s) (mapping) parameter.
    @classmethod
    def start(
        cls,
        measures: Iterable[Measure] = ALL_MEASURES,
        poll_intervall_ms: int = 100,
        system_name: Optional[str] = None,
        system_description: Optional[str] = None,
        export_file_path: Optional[PathLike] = None,
        export_format: Optional[ExportFormat] = None,
    ) -> Self:
        # Get Python info first, and then strip Python measures from the list.
        python_info, measures = _get_python_info(measures=measures, export_file_path=export_file_path)

        # Prepare the measure configurations.
        configs_array = _prepare_measure_configurations(measures)

        # Get other info, first, before starting the tracking.
        result_pointer = pointer(pointer(_Result()))
        error_int = _LIBRARY.tirexFetchInfo(configs_array, result_pointer)
        _handle_error(error_int)
        fetch_info_result = result_pointer.contents

        # Instate the error handler
        error_handle = _init_error_handling()

        # Start the tracking.
        tracking_handle_pointer = pointer(pointer(_TrackingHandle()))
        error_int = _LIBRARY.tirexStartTracking(configs_array, poll_intervall_ms, tracking_handle_pointer)
        _handle_error(error_int)
        tracking_handle = tracking_handle_pointer.contents

        return cls(
            _fetch_info_result=fetch_info_result,
            _tracking_handle=tracking_handle,
            _error_handle=error_handle,
            _python_info=python_info,
            _system_name=system_name,
            _system_description=system_description,
            _export_file_path=export_file_path,
            _export_format=export_format,
            results={},
        )

    def stop(self) -> Mapping[Measure, ResultEntry]:
        result_pointer = pointer(pointer(_Result()))
        error_int = _LIBRARY.tirexStopTracking(self._tracking_handle, result_pointer)
        _handle_error(error_int)

        _deinit_error_handling(self._error_handle)

        self._export(result_pointer.contents)

        self.results.update(_parse_results(self._fetch_info_result))
        self.results.update(self._python_info)
        self.results.update(_parse_results(result_pointer.contents))
        return self.results

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.stop()

    def __len__(self) -> int:
        return len(self.results)

    def __iter__(self) -> Iterator[Measure]:
        return iter(self.results)

    def __getitem__(self, key) -> ResultEntry:
        return self.results[key]

    @overload
    def get(self, key: Measure) -> Optional[ResultEntry]: ...

    @overload
    def get(self, key: Measure, default: T) -> Union[ResultEntry, T]: ...

    def get(self, key: Measure, default: Optional[T] = None) -> Optional[Union[ResultEntry, T]]:  # type: ignore[overload]
        return self.results.get(key, default)

    def __contains__(self, key) -> bool:
        return key in self.results

    def keys(self) -> KeysView[Measure]:
        return self.results.keys()

    def items(self) -> ItemsView[Measure, ResultEntry]:
        return self.results.items()

    def values(self) -> ValuesView[ResultEntry]:
        return self.results.values()

    def __eq__(self, other) -> bool:
        return NotImplemented

    def _export(self, result: "Pointer[_Result]") -> None:
        if self._export_file_path is None:
            return
        elif self._export_format is None:
            self._export_guessed_format(result)
        elif self._export_format == ExportFormat.IR_METADATA:
            self._export_ir_metadata(result)
        else:
            raise ValueError("Invalid export format.")

    def _export_guessed_format(self, result: "Pointer[_Result]") -> None:
        if self._export_file_path is None:
            return
        elif any(
            Path(self._export_file_path).name.endswith(extension)
            for extension in [
                "ir_metadata",
                "ir-metadata",
                "irmetadata",
                ".yml",
                ".yaml",
                "ir_metadata.gz",
                "ir-metadata.gz",
                "irmetadata.gz",
                ".yml.gz",
                ".yaml.gz",
            ]
        ):
            self._export_ir_metadata(result)

    def _export_ir_metadata(self, result: "Pointer[_Result]") -> None:
        if self._export_file_path is None:
            return

        export_file_path = Path(self._export_file_path)
        if export_file_path.exists():
            raise ValueError("Metadata file already exists.")

        # Run the C-internal ir_metadata export.
        _LIBRARY.tirexResultExportIrMetadata(
            self._fetch_info_result,
            result,
            c_char_p(str(export_file_path.resolve()).encode(_ENCODING)),
        )

        # Parse the initial ir_metadata.
        buffer = Path(export_file_path).read_bytes()
        if buffer.startswith(b"ir_metadata.start\n"):
            buffer = buffer[len(b"ir_metadata.start\n") :]
        if buffer.endswith(b"ir_metadata.end\n"):
            buffer = buffer[: -len(b"ir_metadata.end\n")]

        with BytesIO(buffer) as yaml_file:
            tmp_ir_metadata = yaml_safe_load(yaml_file)
        ir_metadata = _recursive_defaultdict()

        # Add user-provided metadata.
        if self._system_name is not None:
            ir_metadata["method"]["name"] = self._system_name
        if self._system_description is not None:
            ir_metadata["method"]["description"] = self._system_description

        # Add Python-specific metadata.
        ir_metadata["implementation"]["executable"]["cmd"] = loads(self._python_info[Measure.PYTHON_EXECUTABLE].value)
        ir_metadata["implementation"]["executable"]["args"] = loads(self._python_info[Measure.PYTHON_ARGUMENTS].value)
        ir_metadata["implementation"]["executable"]["version"] = loads(self._python_info[Measure.PYTHON_VERSION].value)
        ir_metadata["implementation"]["python"]["modules"] = loads(self._python_info[Measure.PYTHON_VERSION].value)
        ir_metadata["implementation"]["python"]["packages"] = loads(
            self._python_info[Measure.PYTHON_INSTALLED_PACKAGES].value
        )
        ir_metadata["implementation"]["python"]["interactive"] = loads(
            self._python_info[Measure.PYTHON_IS_INTERACTIVE].value
        )
        if Measure.PYTHON_SCRIPT_FILE_PATH in self._python_info:
            ir_metadata["implementation"]["script"]["path"] = loads(
                self._python_info[Measure.PYTHON_SCRIPT_FILE_PATH].value
            )
        if Measure.PYTHON_NOTEBOOK_FILE_PATH in self._python_info:
            ir_metadata["implementation"]["notebook"]["path"] = loads(
                self._python_info[Measure.PYTHON_NOTEBOOK_FILE_PATH].value
            )
        ir_metadata["implementation"]["source"]["archive"]["path"] = loads(
            self._python_info[Measure.PYTHON_CODE_ARCHIVE_PATH].value
        )
        ir_metadata["implementation"]["source"]["archive"]["script path"] = loads(
            self._python_info[Measure.PYTHON_SCRIPT_FILE_PATH_IN_CODE_ARCHIVE].value
        )
        if Measure.PYTHON_NOTEBOOK_FILE_PATH_IN_CODE_ARCHIVE in self._python_info:
            ir_metadata["implementation"]["source"]["archive"]["notebook path"] = loads(
                self._python_info[Measure.PYTHON_NOTEBOOK_FILE_PATH_IN_CODE_ARCHIVE].value
            )

        ir_metadata = _deep_merge(ir_metadata, tmp_ir_metadata)
        ir_metadata = _recursive_undefaultdict(ir_metadata)

        global _REGISTERED_METADATA
        if _REGISTERED_METADATA:
            ir_metadata.update(_REGISTERED_METADATA)

        for registered_resolve_to, registered_file, subdir in _REGISTERED_FILES:
            if subdir is None:
                target_file = export_file_path.parent / registered_file
            else:
                target_file = export_file_path.parent / subdir / registered_file
            target_file.parent.mkdir(exist_ok=True, parents=True)
            copy(registered_resolve_to / registered_file, target_file)

        # Serialize the updated ir_metadata.
        file_open: Callable[[], IO[str]]
        if export_file_path.suffix == ".gz":

            def file_open() -> IO[str]:
                return gzip_open(export_file_path, "wt")

        else:

            def file_open() -> IO[str]:
                return export_file_path.open("wt")

        write_prefix_suffix = any(
            Path(self._export_file_path).name.endswith(extension)
            for extension in [
                "ir_metadata",
                "ir-metadata",
                "irmetadata",
                "ir_metadata.gz",
                "ir-metadata.gz",
                "irmetadata.gz",
            ]
        )
        with file_open() as file:
            if write_prefix_suffix:
                file.write("ir_metadata.start\n")

            yaml_safe_dump(
                data=ir_metadata,
                stream=file,
                encoding="utf-8",
            )
            if write_prefix_suffix:
                file.write("ir_metadata.end\n")


# TODO: Add aggregation(s) (mapping) parameter.
def start_tracking(
    measures: Iterable[Measure] = ALL_MEASURES,
    poll_intervall_ms: int = 100,
    system_name: Optional[str] = None,
    system_description: Optional[str] = None,
    export_file_path: Optional[PathLike] = None,
    export_format: Optional[ExportFormat] = None,
) -> TrackingHandle:
    return TrackingHandle.start(
        measures=measures,
        poll_intervall_ms=poll_intervall_ms,
        system_name=system_name,
        system_description=system_description,
        export_file_path=export_file_path,
        export_format=export_format,
    )


def stop_tracking(
    tracking_handle: TrackingHandle,
) -> Mapping[Measure, ResultEntry]:
    return tracking_handle.stop()


# TODO: Add aggregation(s) (mapping) parameter.
def tracking(
    measures: Iterable[Measure] = ALL_MEASURES,
    poll_intervall_ms: int = 100,
    system_name: Optional[str] = None,
    system_description: Optional[str] = None,
    export_file_path: Optional[PathLike] = None,
    export_format: Optional[ExportFormat] = None,
) -> TrackingHandle:
    return TrackingHandle.start(
        measures=measures,
        poll_intervall_ms=poll_intervall_ms,
        system_name=system_name,
        system_description=system_description,
        export_file_path=None if export_file_path is None else Path(export_file_path),
        export_format=export_format,
    )


# TODO: Add aggregation(s) (mapping) parameter.
def track(
    block: Callable[[], None],
    measures: Iterable[Measure] = ALL_MEASURES,
    poll_intervall_ms: int = 100,
    system_name: Optional[str] = None,
    system_description: Optional[str] = None,
    export_file_path: Optional[PathLike] = None,
    export_format: Optional[ExportFormat] = None,
) -> Mapping[Measure, ResultEntry]:
    with tracking(
        measures=measures,
        poll_intervall_ms=poll_intervall_ms,
        system_name=system_name,
        system_description=system_description,
        export_file_path=export_file_path,
        export_format=export_format,
    ) as handle:
        block()
        return handle.results


@overload
def tracked(f_or_measures: Callable[P, T]) -> Union[Callable[P, T], ResultsAccessor]:
    pass


@overload
def tracked(
    f_or_measures: Iterable[Measure] = ALL_MEASURES,
    poll_intervall_ms: int = ...,
    system_name: Optional[str] = ...,
    system_description: Optional[str] = ...,
    export_file_path: Optional[PathLike] = ...,
    export_format: Optional[ExportFormat] = ...,
) -> Callable[[Callable[P, T]], Union[Callable[P, T], ResultsAccessor]]:
    pass


# TODO: Add aggregation(s) (mapping) parameter.
def tracked(
    f_or_measures: Union[Callable[P, T], Iterable[Measure]] = ALL_MEASURES,
    poll_intervall_ms: int = 100,
    system_name: Optional[str] = None,
    system_description: Optional[str] = None,
    export_file_path: Optional[PathLike] = None,
    export_format: Optional[ExportFormat] = None,
) -> Union[
    Union[Callable[P, T], ResultsAccessor],
    Callable[[Callable[P, T]], Union[Callable[P, T], ResultsAccessor]],
]:
    if callable(f_or_measures):
        f = f_or_measures

        results: MutableMapping[Measure, ResultEntry] = {}

        @wraps(f)
        def wrapper(*args, **kwds):
            handle = TrackingHandle.start()
            try:
                return f(*args, **kwds)
            finally:
                tmp_results = handle.stop()
                results.clear()
                results.update(tmp_results)

        results_wrapper = cast(Union[Callable[P, T], ResultsAccessor], wrapper)
        results_wrapper.results = results  # type: ignore
        return results_wrapper

    else:
        measures = f_or_measures

        def decorator(f: Callable[P, T]) -> Union[Callable[P, T], ResultsAccessor]:
            results: MutableMapping[Measure, ResultEntry] = {}

            @wraps(f)
            def wrapper(*args, **kwds):
                handle = TrackingHandle.start(
                    measures=measures,
                    poll_intervall_ms=poll_intervall_ms,
                    system_name=system_name,
                    system_description=system_description,
                    export_file_path=export_file_path,
                    export_format=export_format,
                )
                try:
                    return f(*args, **kwds)
                finally:
                    tmp_results = handle.stop()
                    results.clear()
                    results.update(tmp_results)

            results_wrapper = cast(Union[Callable[P, T], ResultsAccessor], wrapper)
            results_wrapper.results = results  # type: ignore
            return results_wrapper

        return decorator


def clear_metadata_register() -> None:
    global _REGISTERED_METADATA
    _REGISTERED_METADATA = {}


def register_metadata(metadata: Mapping[str, Any]) -> None:
    global _REGISTERED_METADATA
    _REGISTERED_METADATA.update(metadata)


def clear_file_register() -> None:
    global _REGISTERED_FILES
    _REGISTERED_FILES = []


def register_file(resolve_to: Path, file: Path, subdirectory: str = ".") -> None:
    if not resolve_to.is_dir():
        raise ValueError(f"Tirex-tracker resolve_to should point to an directory, got {resolve_to}")
    if not (resolve_to / file).is_file():
        raise ValueError(f"Tirex-tracker resolve_to/file should exist, got {resolve_to / file}")

    _REGISTERED_FILES.append((resolve_to, file, subdirectory))


## Initialize error handling to abort at level CRITICAL
_LIBRARY.tirexSetAbortLevel(LogLevel.CRITICAL)

__all__ = ["ALL_AGGREGATIONS", "ALL_MEASURES", "Aggregation", "Error", "LogLevel", "Measure", "ResultType", "tracking"]



from tirex_tracker import ExportFormat, set_log_callback, tracking

set_log_callback(lambda level, component, msg: print(f"[{level}][{component}] {msg}"))

with tracking(export_format=ExportFormat.IR_METADATA, export_file_path="ir_metadata.yml") as results:
    from time import sleep

    sleep(1)



from enum import IntEnum

ENCODING = "ascii"


class Error(IntEnum):
    SUCCESS = 0
    INVALID_ARGUMENT = 1


class Measure(IntEnum):
    OS_NAME = 0
    OS_KERNEL = 1
    TIME_START = 44
    TIME_STOP = 45
    TIME_ELAPSED_WALL_CLOCK_MS = 2
    TIME_ELAPSED_USER_MS = 3
    TIME_ELAPSED_SYSTEM_MS = 4
    CPU_USED_PROCESS_PERCENT = 5
    CPU_USED_SYSTEM_PERCENT = 6
    CPU_AVAILABLE_SYSTEM_CORES = 7
    CPU_ENERGY_SYSTEM_JOULES = 8
    CPU_FEATURES = 9
    CPU_FREQUENCY_MHZ = 10
    CPU_FREQUENCY_MIN_MHZ = 11
    CPU_FREQUENCY_MAX_MHZ = 12
    CPU_VENDOR_ID = 13
    CPU_BYTE_ORDER = 14
    CPU_ARCHITECTURE = 15
    CPU_MODEL_NAME = 16
    CPU_CORES_PER_SOCKET = 17
    CPU_THREADS_PER_CORE = 18
    CPU_CACHES = 19
    CPU_VIRTUALIZATION = 20
    RAM_USED_PROCESS_KB = 21
    RAM_USED_SYSTEM_MB = 22
    RAM_AVAILABLE_SYSTEM_MB = 23
    RAM_ENERGY_SYSTEM_JOULES = 24
    GPU_SUPPORTED = 25
    GPU_MODEL_NAME = 26
    GPU_AVAILABLE_SYSTEM_CORES = 27  # aka. GPU_NUM_CORES
    GPU_USED_PROCESS_PERCENT = 28
    GPU_USED_SYSTEM_PERCENT = 29
    GPU_VRAM_USED_PROCESS_MB = 30
    GPU_VRAM_USED_SYSTEM_MB = 31
    GPU_VRAM_AVAILABLE_SYSTEM_MB = 32
    GPU_ENERGY_SYSTEM_JOULES = 33
    GIT_IS_REPO = 34
    GIT_HASH = 35
    GIT_LAST_COMMIT_HASH = 36
    GIT_BRANCH = 37
    GIT_BRANCH_UPSTREAM = 38
    GIT_TAGS = 39
    GIT_REMOTE_ORIGIN = 40
    GIT_UNCOMMITTED_CHANGES = 41
    GIT_UNPUSHED_CHANGES = 42
    GIT_UNCHECKED_FILES = 43
    GIT_ROOT = 46
    GIT_ARCHIVE_PATH = 47
    PYTHON_VERSION = 1000
    PYTHON_EXECUTABLE = 1001
    PYTHON_ARGUMENTS = 1002
    PYTHON_MODULES = 1003
    PYTHON_INSTALLED_PACKAGES = 1004
    PYTHON_IS_INTERACTIVE = 1005
    PYTHON_SCRIPT_FILE_PATH = 1006
    # 1007 was used in previous versions of the library.
    PYTHON_NOTEBOOK_FILE_PATH = 1008
    # 1009 was used in previous versions of the library.
    PYTHON_CODE_ARCHIVE_PATH = 1010
    PYTHON_SCRIPT_FILE_PATH_IN_CODE_ARCHIVE = 1011
    PYTHON_NOTEBOOK_FILE_PATH_IN_CODE_ARCHIVE = 1012


INVALID_MEASURE = -1


ALL_MEASURES = set(Measure)


class Aggregation(IntEnum):
    NO = 1 << 0
    MAX = 1 << 1
    MIN = 1 << 2
    MEAN = 1 << 3


INVALID_AGGREGATION = -1


ALL_AGGREGATIONS = set(Aggregation)


class ResultType(IntEnum):
    STRING = 0
    INTEGER = 1
    FLOATING = 2
    BOOLEAN = 3
    WSTRING = 4
    STRING_LIST = STRING | 1 << 7
    INTEGER_LIST = INTEGER | 1 << 7
    FLOATING_LIST = FLOATING | 1 << 7
    BOOLEAN_LIST = BOOLEAN | 1 << 7
    WSTRING_LIST = WSTRING | 1 << 7


class LogLevel(IntEnum):
    TRACE = 0
    DEBUG = 1
    INFO = 2
    WARN = 3
    ERROR = 4
    CRITICAL = 5



from tirex_tracker.utils import is_pyterrier_installed

if is_pyterrier_installed():
    from typing import Final, Iterable, Iterator, Mapping, Optional, Tuple, Union

    from pandas import DataFrame
    from pyterrier import Transformer
    from pyterrier.model import IterDict

    from tirex_tracker import (
        ALL_MEASURES,
        ExportFormat,
        Measure,
        PathLike,
        ResultEntry,
        tracking,
    )

    class TrackedTransformer(Transformer):
        transformer: Final[Transformer]
        measures: Final[Iterable[Measure]]
        poll_intervall_ms: Final[int]
        system_name: Final[Optional[str]]
        system_description: Final[Optional[str]]
        export_file_path: Final[Optional[PathLike]]
        export_format: Final[Optional[ExportFormat]]
        _results: Mapping[Measure, ResultEntry]

        def __init__(
            self,
            transformer: Transformer,
            measures: Iterable[Measure] = ALL_MEASURES,
            poll_interval_ms: int = 100,
            system_name: Optional[str] = None,
            system_description: Optional[str] = None,
            export_file_path: Optional[PathLike] = None,
            export_format: Optional[ExportFormat] = None,
        ) -> None:
            super().__init__()
            self.transformer = transformer
            self.measures = measures
            self.poll_intervall_ms = poll_interval_ms
            self.system_name = system_name
            self.system_description = system_description
            self.export_file_path = export_file_path
            self.export_format = export_format

        @property
        def results(self) -> Mapping[Measure, ResultEntry]:
            return self._results

        def transform(self, inp: DataFrame) -> DataFrame:
            with tracking(
                measures=self.measures,
                poll_intervall_ms=self.poll_intervall_ms,
                system_name=self.system_name,
                system_description=self.system_description,
                export_file_path=self.export_file_path,
                export_format=self.export_format,
            ) as tracking_results:
                results = self.transformer.transform(inp)
            self._results = tracking_results
            return results

        def transform_iter(self, inp: IterDict) -> IterDict:
            with tracking(
                measures=self.measures,
                poll_intervall_ms=self.poll_intervall_ms,
                system_name=self.system_name,
                system_description=self.system_description,
                export_file_path=self.export_file_path,
                export_format=self.export_format,
            ) as tracking_results:
                results = self.transformer.transform_iter(inp)
            self._results = tracking_results
            return results

        def transform_gen(
            self,
            input: DataFrame,
            batch_size=1,
            output_topics=False,
        ) -> Union[Iterator[DataFrame], Iterator[Tuple[DataFrame, DataFrame]]]:
            with tracking(
                measures=self.measures,
                poll_intervall_ms=self.poll_intervall_ms,
                system_name=self.system_name,
                system_description=self.system_description,
                export_file_path=self.export_file_path,
                export_format=self.export_format,
            ) as tracking_results:
                results = self.transformer.transform_gen(
                    input, batch_size=batch_size, output_topics=output_topics
                )
            self._results = tracking_results
            return results



import sys
from ctypes import CDLL, POINTER, Structure, c_char_p, c_int, c_size_t, c_void_p, cdll
from importlib import resources
from json import dumps
from pathlib import Path
from sys import platform
from typing import TYPE_CHECKING, Callable, ContextManager, Mapping, NamedTuple, Optional, cast

from importlib_metadata import PackageNotFoundError, version

if TYPE_CHECKING:
    from ctypes import Array
    from ctypes import _CFunctionType as CFunctionType  # type: ignore
    from ctypes import _Pointer as Pointer  # type: ignore

from .constants import ENCODING, INVALID_AGGREGATION, INVALID_MEASURE, Aggregation, Measure, ResultType


class _TrackingHandle(Structure):
    _fields_ = []


class MeasureConfiguration(NamedTuple):
    measure: Measure
    aggregation: Aggregation


class _MeasureConfiguration(Structure):
    measure: int
    aggregation: int

    _fields_ = [
        ("measure", c_int),
        ("aggregation", c_int),
    ]


NULL_MEASURE_CONFIGURATION = _MeasureConfiguration(INVALID_MEASURE, INVALID_AGGREGATION)


class _Result(Structure):
    _fields_ = []


class ResultEntry(NamedTuple):
    source: Measure
    value: str
    type: ResultType


class _ResultEntry(Structure):
    source: int
    value: int
    type: int

    _fields_ = [
        ("source", c_int),
        ("value", c_void_p),
        ("type", c_int),
    ]


class MeasureInfo(NamedTuple):
    description: str
    data_type: ResultType
    example: str


class _MeasureInfo(Structure):
    description: bytes
    result_type: int
    example: bytes

    _fields_ = [
        ("description", c_char_p),
        ("result_type", c_int),
        ("example", c_char_p),
    ]

    def to_measure_info(self) -> MeasureInfo:
        return MeasureInfo(
            description=self.description.decode(ENCODING),
            data_type=ResultType(self.result_type),
            example=self.example.decode(ENCODING),
        )


_PYTHON_MEASURES: Mapping[Measure, MeasureInfo] = {
    Measure.PYTHON_VERSION: MeasureInfo(
        description="Python version used to run the tracked program.",
        data_type=ResultType.STRING,
        example=dumps("3.12.0"),
    ),
    Measure.PYTHON_EXECUTABLE: MeasureInfo(
        description="Python executable used to run the program.",
        data_type=ResultType.STRING,
        example=dumps("/usr/bin/python3"),
    ),
    Measure.PYTHON_ARGUMENTS: MeasureInfo(
        description="Arguments passed to the Python executable.",
        data_type=ResultType.STRING_LIST,
        example=dumps(["-m", "tirex_tracker"]),
    ),
    Measure.PYTHON_MODULES: MeasureInfo(
        description="Python modules visible in the current environment.",
        data_type=ResultType.STRING_LIST,
        example=dumps(["os", "sys", "json"]),
    ),
    Measure.PYTHON_INSTALLED_PACKAGES: MeasureInfo(
        description="Python packages installed in the current environment.",
        data_type=ResultType.STRING_LIST,
        example=dumps(["tirex-tracker==0.1.0", "numpy==1.21.2"]),
    ),
    Measure.PYTHON_IS_INTERACTIVE: MeasureInfo(
        description="True if the Python interpreter is run interactively.",
        data_type=ResultType.BOOLEAN,
        example=dumps(True),
    ),
    Measure.PYTHON_SCRIPT_FILE_PATH: MeasureInfo(
        description="Path to the Python script file.",
        data_type=ResultType.STRING,
        example=dumps("/path/to/script.py"),
    ),
    Measure.PYTHON_NOTEBOOK_FILE_PATH: MeasureInfo(
        description="Path to the Jupyter notebook file.",
        data_type=ResultType.STRING,
        example=dumps("/path/to/notebook.ipynb"),
    ),
    Measure.PYTHON_CODE_ARCHIVE_PATH: MeasureInfo(
        description="The archive that contains a snapshot of the code.",
        data_type=ResultType.STRING,
        example=dumps("/path/to/code.zip"),
    ),
    Measure.PYTHON_SCRIPT_FILE_PATH_IN_CODE_ARCHIVE: MeasureInfo(
        description="The script that was executed in the code archive.",
        data_type=ResultType.STRING,
        example=dumps("script.py"),
    ),
    Measure.PYTHON_NOTEBOOK_FILE_PATH_IN_CODE_ARCHIVE: MeasureInfo(
        description="The notebook that was executed in the code archive.",
        data_type=ResultType.STRING,
        example=dumps("notebook.ipynb"),
    ),
}


class ProviderInfo(NamedTuple):
    name: str
    description: str
    version: Optional[str]


class _ProviderInfo(Structure):
    name: bytes
    description: bytes
    version: Optional[bytes]

    _fields_ = [
        ("name", c_char_p),
        ("description", c_char_p),
        ("version", c_char_p),
    ]

    def to_provider(self) -> ProviderInfo:
        return ProviderInfo(
            name=self.name.decode(ENCODING),
            description=self.description.decode(ENCODING),
            version=(self.version.decode(ENCODING) if self.version is not None else None),
        )


try:
    provider_version = version("tirex-tracker")
except PackageNotFoundError:
    # if the TIREx tracker is not installed via pip
    provider_version = "unpublished"


class _TirexTrackerLibrary(CDLL):
    tirexResultEntryGetByIndex: Callable[["Pointer[_Result]", c_size_t, "Pointer[_ResultEntry]"], int]
    tirexResultEntryNum: Callable[["Pointer[_Result]", "Pointer[c_size_t]"], int]
    tirexResultFree: Callable[["Pointer[_Result]"], None]
    tirexFetchInfo: Callable[["Array[_MeasureConfiguration]", "Pointer[Pointer[_Result]]"], int]
    tirexStartTracking: Callable[["Array[_MeasureConfiguration]", int, "Pointer[Pointer[_TrackingHandle]]"], int]
    tirexStopTracking: Callable[["Pointer[_TrackingHandle]", "Pointer[Pointer[_Result]]"], int]
    tirexSetLogCallback: Callable[["Optional[CFunctionType]"], None]
    tirexDataProviderGetAll: Callable[["Array[_ProviderInfo]", int], int]
    tirexMeasureInfoGet: Callable[[int, "Pointer[Pointer[_MeasureInfo]]"], int]
    tirexResultExportIrMetadata: Callable[["Pointer[_Result]", "Pointer[_Result]", c_char_p], int]


def __get_library() -> ContextManager[Path]:
    path: str
    if platform == "linux":
        path = "libtirex_tracker.so"
    elif platform == "darwin":
        path = "libtirex_tracker.dylib"
    elif platform == "win32":
        path = "tirex_tracker.dll"
    else:
        raise RuntimeError("Unsupported platform.")
    top_package = __package__.split(".", maxsplit=1)[0]  # Should evaluate to "tirex_tracker"
    if sys.version_info >= (3, 9):  # Python 3.9 or later:
        return resources.as_file(resources.files(top_package).joinpath(path))
    # Python 3.7, Python 3.8: (this is deprecated and should not be used in later versions)
    return resources.path(top_package, path)


def _load_library() -> _TirexTrackerLibrary:
    with __get_library() as lib:
        library = cdll.LoadLibrary(str(lib))
    library.tirexResultEntryGetByIndex.argtypes = [
        POINTER(_Result),
        c_size_t,
        POINTER(_ResultEntry),
    ]
    library.tirexResultEntryGetByIndex.restype = c_int
    library.tirexResultEntryNum.argtypes = [POINTER(_Result), POINTER(c_size_t)]
    library.tirexResultEntryNum.restype = c_int
    library.tirexResultFree.argtypes = [POINTER(_Result)]
    library.tirexResultFree.restype = c_void_p
    library.tirexFetchInfo.argtypes = [
        POINTER(_MeasureConfiguration),
        POINTER(POINTER(_Result)),
    ]
    library.tirexFetchInfo.restype = c_int
    library.tirexStartTracking.argtypes = [
        POINTER(_MeasureConfiguration),
        c_size_t,
        POINTER(POINTER(_TrackingHandle)),
    ]
    library.tirexStartTracking.restype = c_int
    library.tirexStopTracking.argtypes = [
        POINTER(_TrackingHandle),
        POINTER(POINTER(_Result)),
    ]
    library.tirexStopTracking.restype = c_int
    library.tirexSetLogCallback.argtypes = [c_void_p]
    library.tirexSetLogCallback.restype = c_void_p
    library.tirexDataProviderGetAll.argtypes = [POINTER(_ProviderInfo), c_size_t]
    library.tirexDataProviderGetAll.restype = c_size_t
    library.tirexMeasureInfoGet.argtypes = [c_int, POINTER(POINTER(_MeasureInfo))]
    library.tirexMeasureInfoGet.restype = c_int
    library.tirexResultExportIrMetadata.argtypes = [
        POINTER(_Result),
        POINTER(_Result),
        c_char_p,
    ]
    library.tirexResultExportIrMetadata.restype = c_int
    library.tirexSetAbortLevel.argtypes = [c_int]
    library.tirexSetAbortCallback.argtypes = [c_void_p]
    return cast(_TirexTrackerLibrary, library)


LIBRARY = _load_library()



import os
import unittest
from ctypes import pointer
from threading import Event, Thread

from tirex_tracker import (
    _LIBRARY,
    _NULL_MEASURE_CONFIGURATION,
    Aggregation,
    LogLevel,
    _MeasureConfiguration,
    _Result,
    set_log_callback,
)
from tirex_tracker._utils.errorhandling import deinit_error_handling, init_error_handling

set_log_callback(lambda level, component, message: print(f"[{level}][{component}] {message}"))


class TestErrorHandling(unittest.TestCase):
    def test_error_thrown(self):
        _LIBRARY.tirexSetAbortLevel(LogLevel.WARN)
        handle = init_error_handling()
        set_log_callback(lambda level, component, message: print(f"[{level}][{component}] {message}"))

        def should_raise():
            latch_stopped = Event()
            print("Parent:", os.getpid())

            def do_work():
                print("Child:", os.getpid())
                measures = [_MeasureConfiguration(-7, Aggregation.NO.value), _NULL_MEASURE_CONFIGURATION]

                configs = (_MeasureConfiguration * (len(measures)))(*measures)
                resultptr = pointer(_Result())
                _LIBRARY.tirexFetchInfo(configs, pointer(resultptr))
                print(resultptr.contents)
                latch_stopped.set()

            Thread(target=do_work, daemon=True).start()
            latch_stopped.wait()

        self.assertRaises(RuntimeError, should_raise)
        deinit_error_handling(handle)
